{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Pilas\n",
    "\n",
    "## Introducción\n",
    "\n",
    "Las pilas son estructuras de datos fundamentales en la programación que siguen el principio de **LIFO** (Last In, First Out), es decir, el último elemento en entrar es el primero en salir. Esto las hace útiles para resolver problemas en los que la última acción realizada debe ser la primera en revertirse, como en la función \"deshacer\" de los editores de texto o la navegación de un historial de páginas web.\n",
    "\n",
    "En esta clase, exploraremos el concepto de pilas, sus operaciones básicas y cómo implementarlas en Java utilizando la clase `Stack` del paquete `java.util` y cómo podemos implementarlas manualmente. Al final de la clase, aplicaremos estos conceptos para resolver un problema práctico.\n",
    "\n",
    "Algunos ejemplos de pilas son:\n",
    "\n",
    "\n",
    "::::{grid}\n",
    "\n",
    ":::{grid-item}\n",
    ":margin: auto auto 0 0\n",
    ":columns: 6\n",
    "\n",
    "\n",
    "```{figure} ../../images/stack1.png\n",
    "---\n",
    "width: 100%\n",
    "name: stack1\n",
    "---\n",
    "Pila de panqueques.\n",
    "```\n",
    ":::\n",
    "\n",
    ":::{grid-item}\n",
    ":margin: auto auto 0 0\n",
    ":columns: 6\n",
    "\n",
    "```{figure} ../../images/stack2.png\n",
    "---\n",
    "width: 100%\n",
    "name: Stack2\n",
    "---\n",
    "Milhoja.\n",
    "```\n",
    ":::\n",
    "\n",
    ":::{grid-item}\n",
    ":margin: auto auto 0 0\n",
    ":columns: 2\n",
    "\n",
    "```{figure} https://i.etsystatic.com/12707189/r/il/9459c7/3154225373/il_fullxfull.3154225373_kotx.jpg\n",
    "---\n",
    "width: 100%\n",
    "name: pilalibros\n",
    "---\n",
    "Libros.\n",
    "```\n",
    ":::\n",
    "\n",
    ":::{grid-item}\n",
    ":margin: auto auto 0 0\n",
    ":columns: 2\n",
    "\n",
    "```{figure} https://images.fineartamerica.com/images/artworkimages/mediumlarge/1/stack-of-dishes-luis-crump.jpg\n",
    "---\n",
    "width: 100%\n",
    "name: Pilaplatos\n",
    "---\n",
    "Platos.\n",
    "```\n",
    ":::\n",
    "\n",
    ":::{grid-item}\n",
    ":margin: auto auto 0 0\n",
    ":columns: 4\n",
    "\n",
    "```{figure} https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcTDi7Op6u8PtiAU544zDmXE0ct1dZfj1DQFug&s\n",
    "---\n",
    "width: 90%\n",
    "name: PilaDiscos\n",
    "---\n",
    "Discos.\n",
    "```\n",
    "\n",
    ":::\n",
    ":::{grid-item}\n",
    ":margin: auto auto 0 0\n",
    ":columns: 4\n",
    "\n",
    "```{figure} https://www.wikihow.com/images_en/thumb/2/2c/Play-Pogs-Step-6.jpg/v4-460px-Play-Pogs-Step-6.jpg.webp\n",
    "---\n",
    "width: 90%\n",
    "name: pilatazos\n",
    "---\n",
    "Tazos.\n",
    "```\n",
    "\n",
    ":::\n",
    "\n",
    "::::\n",
    "\n",
    "\n",
    "Esta clase abordará los conceptos de las pilas, su implementación en Java, y ejemplos prácticos que ilustran su uso.\n",
    "\n",
    "## Objetivos\n",
    "\n",
    "- Comprender el concepto de pilas y su estructura de datos.\n",
    "- Implementar y utilizar pilas en Java.\n",
    "- Aprender las operaciones básicas de una pila: `push`, `pop`, `peek`, `isEmpty`.\n",
    "- Aplicar pilas en problemas comunes de programación.\n",
    "- Desarrollar una implementación manual de una pila.\n",
    "  \n",
    "## ¿Qué es una Pila?\n",
    "\n",
    "Una pila es una estructura de datos abstracta que permite almacenar y recuperar elementos de manera organizada. En una pila, las operaciones de inserción y eliminación se realizan únicamente en la parte superior. Las principales operaciones de una pila son:\n",
    "\n",
    "1. **push(elemento)**: Inserta un nuevo elemento en la parte superior de la pila.\n",
    "2. **pop()**: Elimina y devuelve el elemento en la parte superior de la pila.\n",
    "3. **peek()**: Devuelve el elemento en la parte superior de la pila sin eliminarlo.\n",
    "4. **isEmpty()**: Verifica si la pila está vacía.\n",
    "\n",
    "### Ejemplo de pila en la vida real\n",
    "Imagina una pila de platos. El último plato que se apila será el primero en retirarse, siguiendo el principio LIFO.\n",
    "\n",
    "## Tipos de Operaciones en una Pila\n",
    "\n",
    "- **push**: Agrega un elemento a la pila.\n",
    "- **pop**: Remueve el elemento en la parte superior de la pila.\n",
    "- **peek**: Obtiene el valor en la parte superior sin removerlo.\n",
    "- **isEmpty**: Comprueba si la pila está vacía.\n",
    "\n",
    "\n",
    "## Implementación de Pilas en Java\n",
    "\n",
    "### Implementación Simple\n",
    "\n",
    "A continuación, se presenta una implementación simple de una pila utilizando una clase genérica en Java."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Clase PilaGenérica\n",
    "public class PilaGenerica<T> {\n",
    "  private Object[] elementos;\n",
    "  private int capacidad;\n",
    "  private int cima;\n",
    "\n",
    "  // Constructor\n",
    "  public PilaGenerica(int tamaño) {\n",
    "      capacidad = tamaño;\n",
    "      elementos = new Object[capacidad];\n",
    "      cima = -1;\n",
    "  }\n",
    "\n",
    "  // Método para agregar un elemento a la pila\n",
    "  public void push(T elemento) {\n",
    "      if (cima == capacidad - 1) {\n",
    "          System.out.println(\"La pila está llena. No se puede agregar el elemento.\");\n",
    "          return;\n",
    "      }\n",
    "      elementos[++cima] = elemento;\n",
    "      System.out.println(\"Elemento \" + elemento + \" agregado a la pila.\");\n",
    "  }\n",
    "\n",
    "  // Método para retirar un elemento de la pila\n",
    "  public T pop() {\n",
    "      if (isEmpty()) {\n",
    "          System.out.println(\"La pila está vacía. No se puede retirar ningún elemento.\");\n",
    "          return null;\n",
    "      }\n",
    "      return (T) elementos[cima--];\n",
    "  }\n",
    "\n",
    "  // Método para obtener el elemento en la cima de la pila sin retirarlo\n",
    "  public T peek() {\n",
    "      if (isEmpty()) {\n",
    "          System.out.println(\"La pila está vacía. No hay elementos en la cima.\");\n",
    "          return null;\n",
    "      }\n",
    "      return (T) elementos[cima];\n",
    "  }\n",
    "\n",
    "  // Método para verificar si la pila está vacía\n",
    "  public boolean isEmpty() {\n",
    "      return cima == -1;\n",
    "  }\n",
    "\n",
    "  // Método para obtener el tamaño de la pila\n",
    "  public int size() {\n",
    "      return cima + 1;\n",
    "  }\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Edita y prueba el código"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Agrega tu código aquí"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Ejemplo de Uso\n",
    "\n",
    "A continuación, se presenta un ejemplo de cómo utilizar la clase `PilaGenerica` para almacenar y manejar una pila de enteros."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Clase Principal para probar la pila\n",
    "public class Principal {\n",
    "  public static void main(String[] args) {\n",
    "      PilaGenerica<Integer> pila = new PilaGenerica<>(5);\n",
    "      \n",
    "      // Agregar elementos a la pila\n",
    "      pila.push(10);\n",
    "      pila.push(20);\n",
    "      pila.push(30);\n",
    "      \n",
    "      // Mostrar el elemento en la cima\n",
    "      System.out.println(\"Elemento en la cima: \" + pila.peek());\n",
    "      \n",
    "      // Retirar un elemento\n",
    "      System.out.println(\"Elemento retirado: \" + pila.pop());\n",
    "      \n",
    "      // Mostrar el tamaño de la pila\n",
    "      System.out.println(\"Tamaño de la pila: \" + pila.size());\n",
    "      \n",
    "      // Retirar todos los elementos\n",
    "      while (!pila.isEmpty()) {\n",
    "          System.out.println(\"Elemento retirado: \" + pila.pop());\n",
    "      }\n",
    "      \n",
    "      // Intentar retirar un elemento de una pila vacía\n",
    "      pila.pop();\n",
    "  }\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Edita y prueba el código"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Agrega tu código aquí"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Implementemos la clase del libro guía:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [],
   "source": [
    "import java.util.ArrayList;\n",
    "public class GenericStack<E> {\n",
    "  private ArrayList<E> list = new ArrayList<>();\n",
    "\n",
    "  public int getSize() {\n",
    "    return list.size();\n",
    "  }\n",
    "\n",
    "  public E peek() {\n",
    "    return list.get(getSize() - 1);\n",
    "  }\n",
    "\n",
    "  public void push(E o) {\n",
    "    list.add(o);\n",
    "  }\n",
    "\n",
    "  public E pop() {\n",
    "    E o = list.get(getSize() - 1);\n",
    "    list.remove(getSize() - 1);\n",
    "    return o;\n",
    "  }\n",
    "\n",
    "  public boolean isEmpty() {\n",
    "    return list.isEmpty();\n",
    "  }\n",
    "  \n",
    "  @Override\n",
    "  public String toString() {\n",
    "    return \"stack: \" + list.toString();\n",
    "  }\n",
    "}\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Crea una clase llamada `TestStack` que cree dos pilas con diferente tipo de elementos, agrega unos archivos, imprime la lista, y elimina elementos de esta."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Agrega tu código aquí"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**¿Qué diferencia hay entre esta clase y la primera?**"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Implementación con Nodos\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Clase Nodo que contiene el dato y una referencia al siguiente nodo\n",
    "class Nodo<T> {\n",
    "  T dato;        // Dato que almacenará el nodo\n",
    "  Nodo<T> siguiente;  // Referencia al siguiente nodo en la pila\n",
    "\n",
    "  // Constructor\n",
    "  public Nodo(T dato) {\n",
    "      this.dato = dato;\n",
    "      this.siguiente = null;\n",
    "  }\n",
    "}\n",
    "\n",
    "// Clase Pila que utiliza nodos\n",
    "public class Pila<T> {\n",
    "  private Nodo<T> cima;  // Referencia al nodo que está en la cima de la pila\n",
    "  private int tamanio;   // Variable para llevar el tamaño de la pila\n",
    "\n",
    "  // Constructor: Inicializa la pila vacía\n",
    "  public Pila() {\n",
    "      this.cima = null;  // La pila comienza vacía, por lo que la cima es nula\n",
    "      this.tamanio = 0;  // Tamaño de la pila es 0\n",
    "  }\n",
    "\n",
    "  // Método para agregar un elemento a la pila (Push)\n",
    "  public void push(T dato) {\n",
    "      Nodo<T> nuevoNodo = new Nodo<>(dato);  // Crear un nuevo nodo con el dato\n",
    "      nuevoNodo.siguiente = cima;  // El siguiente nodo del nuevo es el que era la cima\n",
    "      cima = nuevoNodo;  // Actualizamos la cima para que sea el nuevo nodo\n",
    "      tamanio++;  // Incrementamos el tamaño de la pila\n",
    "  }\n",
    "\n",
    "  // Método para remover y devolver el elemento en la cima de la pila (Pop)\n",
    "  public T pop() {\n",
    "      if (isEmpty()) {\n",
    "          System.out.println(\"Error: La pila está vacía.\");\n",
    "          return null;  // Devuelve null si la pila está vacía\n",
    "      }\n",
    "      T dato = cima.dato;  // Obtener el dato del nodo en la cima\n",
    "      cima = cima.siguiente;  // Mover la cima al siguiente nodo\n",
    "      tamanio--;  // Reducimos el tamaño de la pila\n",
    "      return dato;  // Devolvemos el dato del nodo removido\n",
    "  }\n",
    "\n",
    "  // Método para ver el elemento en la cima sin removerlo (Peek)\n",
    "  public T peek() {\n",
    "      if (isEmpty()) {\n",
    "          System.out.println(\"Error: La pila está vacía.\");\n",
    "          return null;  // Devuelve null si la pila está vacía\n",
    "      }\n",
    "      return cima.dato;  // Devuelve el dato del nodo en la cima\n",
    "  }\n",
    "\n",
    "  // Método para verificar si la pila está vacía\n",
    "  public boolean isEmpty() {\n",
    "      return cima == null;  // Si la cima es null, entonces la pila está vacía\n",
    "  }\n",
    "\n",
    "  // Método que devuelve el tamaño actual de la pila\n",
    "  public int size() {\n",
    "      return tamanio;  // Devuelve el número de elementos en la pila\n",
    "  }\n",
    "\n",
    "  // Método para imprimir el contenido de la pila\n",
    "  @Override\n",
    "  public String toString() {\n",
    "      if (isEmpty()) {\n",
    "          return \"La pila está vacía.\";\n",
    "      }\n",
    "      \n",
    "      StringBuilder sb = new StringBuilder();\n",
    "      Nodo<T> actual = cima;\n",
    "      \n",
    "      while (actual != null) {\n",
    "          sb.append(actual.dato).append(\"\\n\");\n",
    "          actual = actual.siguiente;\n",
    "      }\n",
    "      \n",
    "      return sb.toString();\n",
    "  }\n",
    "}\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Explicación de la clase `Pila`\n",
    "\n",
    "- **Atributos:**\n",
    "  - `cima`: Referencia al nodo en la parte superior de la pila.\n",
    "  - `tamanio`: Contador del número de elementos en la pila.\n",
    "\n",
    "- **Métodos:**\n",
    "  - `push(T dato)`: Añade un nuevo elemento a la cima de la pila.\n",
    "  - `pop()`: Elimina y devuelve el elemento en la cima de la pila. Si la pila está vacía, se muestra un mensaje de error.\n",
    "  - `peek()`: Devuelve el elemento en la cima sin eliminarlo. También muestra un mensaje de error si la pila está vacía.\n",
    "  - `isEmpty()`: Devuelve `true` si la pila está vacía, y `false` en caso contrario.\n",
    "  - `size()`: Devuelve el número de elementos en la pila.\n",
    "  - `toString()`: Recorre la pila y crea una representación en cadena de sus elementos.\n",
    "\n",
    "\n",
    "#### Ejemplo"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Aquí el código de prueba"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Aplicaciones\n",
    "\n",
    "Las pilas tienen diversas aplicaciones en la programación, algunas de las más comunes son:\n",
    "\n",
    "1. **Deshacer y Rehacer**: Las aplicaciones que permiten \"deshacer\" (undo) y \"rehacer\" (redo) utilizan pilas para almacenar las acciones del usuario.\n",
    "2. **Evaluación de expresiones aritméticas**: Al evaluar expresiones aritméticas o convertir notaciones infijas a postfijas, las pilas se utilizan para gestionar los operadores y operandos.\n",
    "3. **Recorrido en profundidad**: En el recorrido de estructuras como árboles o grafos, se puede utilizar una pila para implementar la búsqueda en profundidad (DFS).\n",
    "4. **Comprobación de paréntesis balanceados**: Al analizar expresiones, se utiliza una pila para verificar si los paréntesis están correctamente balanceados.\n",
    "\n",
    "### Otras Aplicaciones\n",
    "\n",
    "- Reversa de cadenas.\n",
    "- Palíndromes.\n",
    "- Historial de navegación en navegadores web.\n",
    "- Evaluación de programas en lenguajes de programación.\n",
    "- Conversión de expresiones.\n",
    "- Resolución de laberintos.\n",
    "- Gestión de la memoria.\n",
    "- Recursividad.\n",
    "- Seguimiento de llamadas a funciones.\n",
    "- Servicio de interrupciones de hardware.\n",
    "- Resolución de problemas combinatorios mediante backtracking."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Pilas de Java\n",
    "\n",
    "Ahora revisemos una implementación más completa y robusta para una pila. Para ello, utilizaremos la clase de vector:\n",
    "```{figure} ../../images/Figure20.10.png\n",
    "---\n",
    "width: 90%\n",
    "name: Figure20.10\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n",
    "Con esta clase en mente, ahora definiremos las pilas heredando algunos métodos y atributos de la clase `Vector`.\n",
    "\n",
    "```{figure} ../../images/Figure20.11.png\n",
    "---\n",
    "width: 90%\n",
    "name: Figure20.11\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n",
    "Ahora, implementa un código que pruebe la clase pilas que viene con java, [Stack.java](https://github.com/openjdk/jdk/blob/master/src/java.base/share/classes/java/util/Stack.java). Creen dos pilas con diferentes tipos de objetos."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "// Agrega tu código aquí"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Ejercicio Preparatorio\n",
    "\n",
    "::::{admonition}\n",
    "\n",
    "Para practicar para el examen, resolvamos el siguiente problema: [Matching Brackets](https://www.codeabbey.com/index/task_view/matching-brackets).\n",
    "\n",
    "::::"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Conclusiones\n",
    "\n",
    "Las pilas son una estructura de datos fundamental que permite resolver problemas en los que el último elemento en agregarse es el primero en ser procesado. Al comprender su funcionamiento, podemos aplicarlas en la solución de diversos problemas que involucran la reversión de acciones, la evaluación de expresiones y el manejo de recursividad. En Java, la clase `Stack` del paquete `java.util` simplifica su implementación, pero también es importante entender cómo implementar pilas manualmente para aprender sobre la gestión de memoria y el control de datos en estructuras más complejas.\n",
    "\n",
    "\n",
    "## Recursos Adicionales\n",
    "\n",
    "### Tutoriales y Guías\n",
    "\n",
    "- [Quick Guide to the Java Stack - Baeldung](https://www.baeldung.com/java-stack)\n",
    "- [Java Stack - CodeGym](https://codegym.cc/groups/posts/java-stack)\n",
    "- [Java Stack - Javatpoint](https://www.javatpoint.com/java-stack)\n",
    "- [14. Stacks and Queues](https://www.cs.princeton.edu/courses/archive/fall14/cos126/lectures/14-StacksQueues-2x2.pdf)\n",
    "- [Lecture 2: Stacks and Queues](https://courses.cs.washington.edu/courses/cse373/19su/files/lectures/slides/lecture02.pdf)\n",
    "\n",
    "### Videos\n",
    "\n",
    "- [Java Stack ](https://www.youtube.com/watch?v=13uQ8dgqbno)\n",
    "- [Tutorial de Pilas (Stacks) en Java ](https://www.youtube.com/watch?v=1IxIZ9qXEgw)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Java",
   "language": "java",
   "name": "java"
  },
  "language_info": {
   "codemirror_mode": "java",
   "file_extension": ".jshell",
   "mimetype": "text/x-java-source",
   "name": "java",
   "pygments_lexer": "java",
   "version": "21.0.4+7-Ubuntu-1ubuntu224.04"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
